/*
 * (c) Jan Ciger February 2015
 *
 * GNU Public license
 *
 * Based on the code by:
 * Richard Visokey AD7C - www.ad7c.com
 */

#include <rotary.h>
#include <LiquidCrystal.h>
#include <EEPROM.h>
#include <Bounce2.h>

//Setup some items
#define BTN_A A4
#define BTN_B A5
#define BTN_ENCODER A0

#define W_CLK 8   // Pin 8 - connect to AD9850 module word load clock pin (CLK)
#define FQ_UD 9   // Pin 9 - connect to freq update pin (FQ)
#define DATA 10   // Pin 10 - connect to serial data load pin (DATA)
#define RESET 11  // Pin 11 - connect to reset pin (RST)
#define pulseHigh(pin) {digitalWrite(pin, HIGH); digitalWrite(pin, LOW); }

#define TUNING_VFO 0
#define TUNING_CURSOR 1
#define TUNING_IF 2

// function prototypes
void updateLCD();
void sendFrequency(double frequency);
void tfr_byte(byte data);
void storeFreq();
void loadFreq();
void clampFreq();

// globals
Rotary encoder = Rotary(2, 3);         // Sets the pins the rotary encoder uses.  Must be interrupt pins.
LiquidCrystal lcd(12, 13, 7, 6, 5, 4); // LCD pins

Bounce btnA, btnB, btnEnc;

uint_fast32_t if_freq = 455000;  // IF in Hz - how much to subtract from the VFO freq
int_fast32_t freq = 7200000;     // Starting frequency of VFO
bool freq_dirty = true;
bool display_dirty = true;
bool memory_dirty = true;
bool editing_increment = false;
bool if_active = false;
bool if_additive = false;

uint_fast32_t last_freq_change = millis();

byte f_cursor = 0;
byte f_increment_idx;
byte tuning_mode = TUNING_VFO;

const byte N_INCREMENTS = 10;
unsigned long tuning_increments[] = { 10, 50, 100, 500, 1000, 2500, 5000, 10000, 100000, 1000000 };

void setup()
{
    //Serial.begin(38400);

    // button pin setup
    pinMode(BTN_ENCODER, INPUT_PULLUP); // Connect to a button that goes to GND on push
    pinMode(BTN_A, INPUT_PULLUP);
    pinMode(BTN_B, INPUT_PULLUP);
    digitalWrite(BTN_ENCODER, HIGH); // enable pull-ups
    digitalWrite(BTN_A, HIGH);
    digitalWrite(BTN_B, HIGH);

    // configure debouncers
    btnA.attach(BTN_A);
    btnB.attach(BTN_B);
    btnEnc.attach(BTN_ENCODER);

    btnA.interval(50);
    btnB.interval(50);
    btnEnc.interval(50);

    // lcd setup
    lcd.begin(16, 2);
    updateLCD();

    // encoder interrupts
    PCICR |= (1 << PCIE2);
    PCMSK2 |= (1 << PCINT18) | (1 << PCINT19);
    sei();

    // AD9850 setup
    pinMode(FQ_UD, OUTPUT);
    pinMode(W_CLK, OUTPUT);
    pinMode(DATA, OUTPUT);
    pinMode(RESET, OUTPUT);

    pulseHigh(RESET);
    pulseHigh(W_CLK);
    pulseHigh(FQ_UD);  // this pulse enables serial mode on the AD9850 - Datasheet page 12.

    tuning_mode = TUNING_VFO;
    loadFreq();
    clampFreq();
    
    display_dirty = true;
    freq_dirty = true;
}

void loop()
{
    btnA.update();
    btnB.update();
    btnEnc.update();

    // BTN A pressed, change to cursor mode from VFO mode, otherwise only refresh
    if(btnA.fell())
    {
        if(tuning_mode == TUNING_VFO)
            tuning_mode = TUNING_CURSOR;

        display_dirty = true;
    }

    // BTN A released, force refresh (cursor switching)
    if(btnA.rose())
    {
        display_dirty = true; // Forces refresh to disable the blinking cursor,
                              // but we don't switch mode - sometimes this gets triggered at start
                              // causing unwanted switch to cursor mode
    }

    // BTN B pressed and BTN A released, toggle IF mode
    if(btnB.fell() && btnA.read() == HIGH)
    {
        if_active = !if_active;

        // sanitize the frequency, so that we don't attempt to generate
        // lower freq. than 1MHz and higher than 30MHz
        if(if_active)
            clampFreq();

        display_dirty = true;
        freq_dirty = true;
    }

    // if both BTN A and BTN B are pressed, switch to IF setting mode
    if((btnB.read() == LOW && btnA.fell()) ||
       (btnB.fell() && btnA.read() == LOW))
    {
        if(tuning_mode == TUNING_IF)
        {
            // we are in IF mode already, toggle the IF polarity
            if_additive = !if_additive;
        } else         
            tuning_mode = TUNING_IF;        
            
        display_dirty = true;
        freq_dirty = true;
    }

    // pressing the encoder goes back to VFO mode if not there already
    if(btnEnc.fell())
    {
        // toggle the increment editing if in VFO mode
        if(tuning_mode != TUNING_VFO)
            editing_increment = false;
        else
            editing_increment = !editing_increment;

        tuning_mode = TUNING_VFO;
        display_dirty = true;
    }

    // refresh the display if out of date
    if(display_dirty)
    {
        display_dirty = false;
        updateLCD();
    }

    // update the frequency if out of date
    if(freq_dirty)
    {
        freq_dirty = false;
        sendFrequency(freq);

        last_freq_change = millis();
        memory_dirty = true;
    }

    // write the changed frequency to EEPROM after 2s
    if(memory_dirty && last_freq_change + 2000 < millis())
    {
        storeFreq();
        memory_dirty = false;
    }
}

ISR(PCINT2_vect)
{
    unsigned char dir = encoder.process();
    if(dir == DIR_NONE)
        return;

    switch(tuning_mode)
    {
        case TUNING_CURSOR:
            // btn A held down, move the cursor
            if(btnA.read() == LOW)
            {
                if(dir == DIR_CCW)
                    f_cursor = (f_cursor < 6) ? f_cursor + 1 : f_cursor;
                else
                    f_cursor = (f_cursor > 0) ? f_cursor - 1 : f_cursor;

                display_dirty = true;

            } else {
                // btn A released, change the selected decade

                int_fast32_t delta = 1;
                for(byte i = 0; i < f_cursor; ++i)
                    delta *= 10;
                
                if(dir == DIR_CCW)
                    freq -= delta;
                else
                    freq += delta;

                // clamp the frequency to not exceed the DDS limits
                clampFreq();

                display_dirty = true;
                freq_dirty = true;
            }
            break;

        case TUNING_VFO:
            // change the tuning increment
            if(editing_increment)
            {
                if(dir == DIR_CCW)
                    f_increment_idx = (f_increment_idx > 0) ? f_increment_idx - 1 : 0;
                else
                    f_increment_idx = (f_increment_idx < N_INCREMENTS - 2) ? f_increment_idx + 1 : N_INCREMENTS - 1;

                display_dirty = true;

            } else {
                // tune by the selected increment
                int_fast32_t delta = tuning_increments[f_increment_idx];
                
                if(dir == DIR_CCW)
                    freq -= delta;
                else
                    freq += delta;
                
                // clamp the frequency to not exceed the DDS limits
                clampFreq();

                display_dirty = true;
                freq_dirty = true;
            }
            break;

        case TUNING_IF:
            // btn A held down, move the cursor
            if(btnA.read() == LOW)
            {
                if(dir == DIR_CCW)
                    f_cursor = (f_cursor < 6) ? f_cursor + 1 : f_cursor;
                else
                    f_cursor = (f_cursor > 0) ? f_cursor - 1 : f_cursor;

                display_dirty = true;

            } else {
                // btn A released, change the selected decade

                uint_fast32_t delta = 1;
                for(byte i = 0; i < f_cursor; ++i)
                    delta *= 10;

                // clamp the frequency to sensible limits for the IF
                if(dir == DIR_CW)
                    if_freq = (if_freq + delta < 30e6) ? if_freq + delta : 30e6;
                else
                    if_freq = (if_freq > delta) ? if_freq - delta : 0;
                
                clampFreq();

                display_dirty = true;
                freq_dirty = true;
            }
            break;
    }
}

// frequency calc from datasheet page 8 = <sys clock> * <frequency tuning word>/2^32
void sendFrequency(double frequency)
{
    if(if_active)
    {
        if(if_additive)
            frequency = frequency + if_freq;
        else
            frequency = frequency - if_freq;
    }

    uint32_t freq = frequency * 4294967295u / 125000000u; // note 125 MHz clock on 9850.  You can make 'slight' tuning variations here by adjusting the clock frequency.
    for (int b = 0; b < 4; b++, freq >>= 8)
        tfr_byte(freq & 0xFF);

    tfr_byte(0x000);   // Final control byte, all 0 for 9850 chip
    pulseHigh(FQ_UD);  // Done!  Should see output

    //Serial.println(frequency);
}

// transfers a byte, a bit at a time, LSB first to the 9850 via serial DATA line
void tfr_byte(byte data)
{
    for (int i = 0; i < 8; i++, data >>= 1)
    {
        digitalWrite(DATA, data & 0x01);
        pulseHigh(W_CLK);   //after each bit sent, CLK is pulsed high
    }
}

void updateLCD()
{
    // clear display
    lcd.clear();
    lcd.noCursor();
    lcd.noBlink();

    // convert frequency to string
    String f((tuning_mode == TUNING_IF) ? if_freq : freq);

    // pad if needed
    byte len = f.length();
    String padding;
    for(int i = 0; i < 7 - len; ++i)
        padding += String("0");
    f = padding + f;

    // split the frequency string
    String k,m,r;
    k = f.substring(f.length() - 3);
    m = f.substring(f.length() - 6, f.length() - 3);
    r = f.substring(0, f.length() - 6);

    // pad
    lcd.setCursor(0, 0);
    if(r.length() <= 1)
        lcd.print(" ");

    // print the frequency
    lcd.print(r);
    lcd.print('.');
    lcd.print(m);
    lcd.print('.');
    lcd.print(k);
    lcd.print(" Hz");

    // if IF is active, indicate it
    if(if_active && tuning_mode != TUNING_IF)
    {       
        lcd.print(if_additive ? "+":"-");
        lcd.print("IF");
    }

    // indicate tuning mode
    lcd.setCursor(0, 1);
    if(tuning_mode == TUNING_CURSOR)
        lcd.print("CUR");
    else if(tuning_mode == TUNING_VFO)
        lcd.print("VFO");
    else
    {
        lcd.print("IF ");
        lcd.print(if_additive ? "(F + IF)" : "(F - IF)");
    }

    // in VFO mode display increment in sensible scale
    if(tuning_mode == TUNING_VFO)
    {
        float inc = tuning_increments[f_increment_idx];
        String unit = "";

        if(inc >= 1e6)
        {
            inc /= 1e6;
            unit = " MHz";
        }

        else if(inc >= 1e3)
        {
            inc /= 1e3;
            unit = " kHz";
        }

        else
            unit = " Hz";

        lcd.print("  ");
        lcd.print(inc, 1);
        lcd.print(unit);

        if(editing_increment)
        {
            lcd.setCursor(15, 1);
            lcd.blink();
        } else
            lcd.noBlink();

    }

    // in cursor mode display cursor/blinking square
    else {
        byte c = 9 - f_cursor;
        if(f_cursor >= 3)
            --c;
        if(f_cursor >= 6)
            --c;

        lcd.setCursor(c, 0);
        if(btnA.read() == LOW)
        {
            lcd.noCursor();
            lcd.blink();
        } else {
            lcd.noBlink();
            lcd.cursor();
        }
    }
}

void clampFreq()
{
    if(if_active)
    {
        if(if_additive && freq + if_freq > 30e6)
            freq = 30e6 - if_freq;
        
        if(!if_additive && (freq - if_freq < 1e6 || if_freq >= freq))
            freq = 1e6 + if_freq;
    } else
    {
        if(freq > 30e6)
            freq = 30e6;
        
        if(freq < 1e6)
            freq = 1e6;
    }
}

void storeFreq()
{
    byte *f = (byte *)(&freq);
    EEPROM.write(0, f[0]);
    EEPROM.write(1, f[1]);
    EEPROM.write(2, f[2]);
    EEPROM.write(3, f[3]);

    f = (byte *)(&if_freq);
    EEPROM.write(4, f[0]);
    EEPROM.write(5, f[1]);
    EEPROM.write(6, f[2]);
    EEPROM.write(7, f[3]);

    EEPROM.write(8, if_additive ? 1 : 0);
    EEPROM.write(9, if_active ? 1 : 0);
}

void loadFreq()
{
    // load frequency
    uint_fast32_t f[4];
    f[0] = EEPROM.read(0);
    f[1] = EEPROM.read(1);
    f[2] = EEPROM.read(2);
    f[3] = EEPROM.read(3);

    freq =  f[0] +
           (f[1] << 8) +
           (f[2] << 16) +
           (f[3] << 24);

    // load IF
    f[0] = EEPROM.read(4);
    f[1] = EEPROM.read(5);
    f[2] = EEPROM.read(6);
    f[3] = EEPROM.read(7);

    if_freq =  f[0] +
              (f[1] << 8) +
              (f[2] << 16) +
              (f[3] << 24);

    // load IF polarity
    byte pol = EEPROM.read(8);
    if_additive = (pol > 0);

    // load IF state
    byte if_on = EEPROM.read(9);
    if_active = (if_on > 0);
}
